#!/usr/bin/python3
import sys
import json
import time
import re
import dateutil
import datetime
from dateutil.parser import *
from dateutil.relativedelta import *
from datetime import timedelta
import statistics
import numpy as np
np.seterr(divide='ignore',invalid='ignore')
from astropy.table import Table
from astropy.io import ascii
from scipy import stats
import math
import os.path
import pickle
from Preproc import *
from Trace import *
from Functions import *


class Algorithms:
  def __init__(self,fn,paths=None):
    self.fn = fn
    if os.path.isfile(fn+"_pre.p"):
      self.preproc = pickle.load(open(fn+"_pre.p","rb"))
    else:
      if paths is not None:
        self.preproc = Preproc(fn,paths)
    if os.path.isfile(fn+"_mt.p"):
      self.mt = pickle.load(open(fn+"_mt.p","rb"))
    else:
      self.mt  = self.preproc.merge_traces()


##Algo1###########################################
  def algo1(self):
    temp = {}
    for logname,traces in self.preproc.logs.items():
      for tracename,trace in traces.items():
        for event in trace.events:
          if event.attrib['time:timestamp'] not in temp:
            temp[event.attrib['time:timestamp']]=[]
          temp[event.attrib['time:timestamp']].append(event)
    ret = {}
    for ts,events in temp.items():
      if len(events)>1:
        ret[ts]=events
    write_results(self.fn+"_algo1",ret)

##Algo2#############################################
  def algo2(self):
    algo2 = {'detectSimilarExecutionTime':{},'detectExecutionLimitation':{},'detectDataLimitation':{}}
    user_reg = re.compile('org:.*')
    simextime = {}
    datakeys = {}
    for trname in self.mt.keys():
      for idx1,ev1 in enumerate(self.mt[trname]):
        for k in ev1.attrib.keys():
          if user_reg.match(k) is not None:
            if ev1.attrib[k] not in datakeys:
              datakeys[ev1.attrib[k]]={}
            if ev1.attrib['concept:name'] not in datakeys[ev1.attrib[k]]:
              datakeys[ev1.attrib[k]][ev1.attrib['concept:name']] = []
            datakeys[ev1.attrib[k]][ev1.attrib['concept:name']].append(ev1)
        for idx2 in range(idx1+1,len(self.mt[trname])):
          ev2 = self.mt[trname][idx2]
          if ev1.attrib['concept:name'] == ev2.attrib['concept:name'] and ev1.attrib['lifecycle:transition'] == "start" and ev2.attrib['lifecycle:transition'] == "complete":
            name = ev1.attrib['concept:name']
            delta = (parse(ev2.attrib['time:timestamp']) - parse(ev1.attrib['time:timestamp'])).total_seconds()
            if name not in simextime:
              simextime[name]={}
            if delta not in simextime[name]:
              simextime[name][delta]=[]
            simextime[name][delta].append(EventPair(ev1,ev2))
            break
    for k,v in datakeys.items():
      algo2['detectExecutionLimitation'][k] = v
    for resources,event_labels in algo2['detectExecutionLimitation'].items():
      algo2['detectDataLimitation'][resources] = {}
      for event_label,events in event_labels.items():
        algo2['detectDataLimitation'][resources][event_label]={'events':events}
        value_list = {}
        for event in events:
          for key,value in event.attrib.items():
            if key == "concept:name" or key == "lifecycle:transition" or key == "time:timestamp" or key == "org:resource":
              continue
            elif value.isnumeric():
              n = int(value)
              if key not in value_list:
                value_list[key]=[]
              value_list[key].append(n)
        for key,values in value_list.items():
          algo2['detectDataLimitation'][resources][event_label][key] = {'average':sum(values)/float(len(values)),'max':max(values)}


    for label,delta in simextime.items():
      sum_of_deltas = 0
      for deltavalue,evps in delta.items():
        sum_of_deltas += len(evps)
      for deltavalue,evps in delta.items():
        percentage = len(evps)*100/sum_of_deltas
        if label not in algo2['detectSimilarExecutionTime']:
          algo2['detectSimilarExecutionTime'][label]={}
        algo2['detectSimilarExecutionTime'][label][repr(percentage)+"%_"+repr(deltavalue)]=evps

    write_results(self.fn+"_algo2",algo2)

##Algo3################################################
  def algo3(self):
    print("algo3")
    ret = {}
    if len(self.preproc.logs.keys()) < 2 or not(self.preproc.lifecycle_options['lifecycle_exists']==1 and "start" in self.preproc.lifecycle_options['lifecycle_options']):
      raise RuntimeError("Algorithm 3 only works with at least 2 log files and without any lifecycles or at least lifecycle transition start")
    for trname,events in self.mt.items():
      for idx,ev1 in enumerate(events):
        if self.preproc.lifecycle_options['lifecycle_exists']==1 and ev1.attrib['lifecycle:transition']!='start':
          continue
        idx_new = 1
        if(idx+idx_new != len(events)):
          ev2 = events[idx+idx_new]
        else:
          break
        breaky = 0
        while(not (self.preproc.lifecycle_options['lifecycle_exists']==1 and ev2.attrib['lifecycle:transition']=='start' or self.preproc.lifecycle_options['lifecycle_exists']==0) or ev1.attrib['time:timestamp'] == ev2.attrib['time:timestamp'] ):
          idx_new += 1
          if(idx_new+idx == len(events)):
            breaky=1
            break
          ev2 = events[idx+idx_new]
        if breaky == 1:
          break
        if ev1.log != ev2.log:
          name = [ev1.attrib['concept:name'],ev2.attrib['concept:name']]
          name = str.join("/",name)
          if name not in ret:
            ret[name]=[EventPair(ev1,ev2)]
          else:
            ret[name].append(EventPair(ev1,ev2))
    write_results(self.fn+"_algo3",ret)

##Algo4##################################################
  def algo4(self):
    ret_pairs = {}
    if len(self.preproc.logs.keys()) < 2 or not(self.preproc.lifecycle_options['lifecycle_exists']==1 and "start" in self.preproc.lifecycle_options['lifecycle_options'] and "complete" in self.preproc.lifecycle_options['lifecycle_options']):
      raise RuntimeError("Algorithm 4 only works with at least 2 log files and lifecycle transitions start and complete are mandatory")
    for trname,events in self.mt.items():
        for idx1,ev1 in enumerate(events):
          for idx2 in range(idx1+1,len(events)):
            ev2 = events[idx2]
            if ev1.attrib['concept:name'] == ev2.attrib['concept:name'] and ev1.attrib['lifecycle:transition'] == "start" and ev2.attrib['lifecycle:transition'] == "complete":
              if trname in ret_pairs:
                ret_pairs[trname].append(EventPair(ev1,ev2))
              else:
                ret_pairs[trname]=[EventPair(ev1,ev2)]
              break
    ret = {}
    for trname,events in ret_pairs.items():
      for idx1,ev1 in enumerate(events):
        for idx2 in range(idx1+1,len(events)):
          ev2 = events[idx2]
          if parse(ev2.ev1.attrib['time:timestamp']) < parse(ev1.ev1.attrib['time:timestamp']):
            event1 = events[idx2]
            event2 = events[idx1]
          else:
            event1 = events[idx1]
            event2 = events[idx2]

          if event1.ev2.log != event2.ev1.log and max(parse(event1.ev2.attrib['time:timestamp']), parse(event2.ev1.attrib['time:timestamp'])) == parse(event2.ev1.attrib['time:timestamp']):
             name = [event1.ev2.attrib['concept:name'],event2.ev1.attrib['concept:name']]
             name = str.join("/",name)
             if name not in ret:
               ret[name]=[{'ev1':{'ev1start':event1.ev1,'ev1end':event1.ev2},'ev2' : {'ev2start':event2.ev1,'ev2end':event2.ev2}}]
             else:
              ret[name].append({'ev1':{'ev1start':event1.ev1,'ev1end':event1.ev2},'ev2' : {'ev2start':event2.ev1,'ev2end':event2.ev2}})
    write_results(self.fn+"_algo4",ret)


