import json
import os.path
from Preproc import *

#is needed for writing JSON files in the desired format
class SetEncoder(json.JSONEncoder):# {{{
  def default(self, obj):
    if isinstance(obj, set):
      return list(obj)
    if isinstance(obj,Event):
      return {"log" : obj.log,"trace" : obj.trace, "attrib" :  obj.attrib}
    if isinstance(obj,EventPair):
      return {'ev1':obj.ev1,'ev2':obj.ev2}
    if isinstance(obj,DeltaTs):
      return {'ev1':obj.ev1, 'ev2':obj.ev2,'delta':obj.delta}
    return json.JSONEncoder.default(self, obj)# }}}

#class OneHundredPercent; determines the absolute numbers of occurrences of events
class OneHundredPercent:# {{{
    def __init__(self,traces):
      self.traces = {}
      if isinstance(traces,dict):
        for tridentifier,tr in traces.items():
          self.traces[tridentifier] = {}
          for event in tr:
            if event.attrib['concept:name'] in self.traces[tridentifier]:
              self.traces[tridentifier][event.attrib['concept:name']]=self.traces[tridentifier][event.attrib['concept:name']]+1
            else:
              self.traces[tridentifier][event.attrib['concept:name']]=1
      if isinstance(traces,list):
        for trace in traces:
          for event in trace:
            if event.trace in self.traces:
              if event.attrib['concept:name'] in self.traces[event.trace]:
                self.traces[event.trace][event.attrib['concept:name']]= self.traces[event.trace][event.attrib['concept:name']]+1
              else:
                self.traces[event.trace][event.attrib['concept:name']]= 1
            else:
              self.traces[event.trace]={}
              self.traces[event.trace][event.attrib['concept:name']]= 1
    def __repr__(self):
      return("%s\n" %(self.traces))

#select attribute; unique identifier is displayed as recommendation
def select_attrib(attribs):# {{{
  if isinstance(attribs, set):
    attribs = list(attribs)
  if len(attribs)==1:
    return attribs[0]
  print("Select an attribute on which traces should be merged from the following list: ")
  for count, val in enumerate(attribs):
    print(count, val)
  inp = input()
  return attribs[int(inp)]
# }}}

#write results to file
def write_results(fn,data):# {{{
  path = os.path.join('results/',fn)
  with open(path+".json",'w') as f:
    f.write(json.dumps(data,cls=SetEncoder,sort_keys=True,indent=4))
  f.closed
# }}}



