#!/usr/bin/python3
import xml.sax as SX
import sys
import json
import time
import dateutil
import datetime
from dateutil.parser import *
from dateutil.relativedelta import *
from datetime import timedelta
import statistics
import numpy as np
np.seterr(divide='ignore',invalid='ignore')
from astropy.table import Table
from astropy.io import ascii
from scipy import stats
import math
import os.path
import pickle
from Trace import *
from Functions import *

#reads the .xes file(s)
class EHandler (SX.ContentHandler):# {{{
  def __init__(self,path,logs,lifecycle_options):
    self.logs = logs
    self.log = path
    self.lifecycle_options = lifecycle_options
    self.tag = None
    self.attrib = None
    self.tracename = None
    self.trace = []
    self.mode = None
  def startElement(self,tag,attributes):
    self.tag = tag
    if tag == "trace":
      self.mode = "trace"
    elif tag == "event":
      self.mode ="event"
      self.attrib={}
    elif tag == "string" and self.mode=="trace" and attributes['key']=="concept:name":
      self.tracename = attributes['value']
    elif self.mode=="event" and tag != "event" and tag != "trace" and tag != "log":
      self.attrib[attributes['key']]=attributes['value']
  def endElement(self,tag):
    if tag == "trace":
      if self.log in self.logs:
        if self.tracename in self.logs[self.log]:
          raise RuntimeError("Trace already exists")
      else:
        self.logs[self.log] = {}
      self.logs[self.log][self.tracename] = Trace(self.log,self.tracename,self.trace)
      self.tracename = None
      self.trace = []
    if tag == "event":
      if 'lifecycle:transition' not in self.attrib.keys():
        self.lifecycle_options['lifecycles_exists'] = 0  # if there is at least one event without a lifecycle transition, then we cannot use lifecycles.
      else:
        self.lifecycle_options['lifecycle_options'].add(self.attrib['lifecycle:transition'])
      self.trace.append(Event(self.log,self.tracename,self.attrib))
      self.attrib = None
    self.tag = tag
# }}}

#preprocessing
class Preproc:
  def __init__(self,fn,paths):
    self.lifecycle_options = {'lifecycle_exists':1,'lifecycle_options':set()}
    self.logs = {}
    self.fn = fn

    for path in paths:
      parser = SX.make_parser()
      handler = EHandler(path,self.logs,self.lifecycle_options)
      parser.setContentHandler(handler)
      parser.parse(path)

    pickle.dump(self,open(fn+"_pre.p","wb"))

  def get_attribset(self):
    ret = set()
    for logname,traces in self.logs.items():
      for tracename,trace in traces.items():
        for k,v in trace.attribset.items():
          if len(v)==1 and k!="concept:instance":
            ret.add(k)
    return ret
  def merge_traces(self):
    ret = {}  #inp => [events]
    if len(self.logs.keys())== 1:
      for logname,traces in self.logs.items():
        for tracename,trace in traces.items():
          ret[tracename]=trace.events
      pickle.dump(ret,open(self.fn+"_mt.p","wb"))
      return ret
    inp = select_attrib(self.get_attribset())
    for logname,traces in self.logs.items():
      for tracename,trace in traces.items():
        if inp in trace.attribset.keys():
          if len(trace.attribset[inp])>1:
            raise RuntimeError("Only unique trace identifiers possible for merging.")
          temp = list(trace.attribset[inp])
          if temp[0] not in ret:
            ret[temp[0]]=[]
          ret[temp[0]].extend(trace.events)
    for trname,events in ret.items():
      events.sort(key=lambda x: parse(x.attrib['time:timestamp']))
    pickle.dump(ret,open(self.fn+"_mt.p","wb"))
    return ret
